import pandas as pd
import numpy as np
import warnings


class Simulation:
    def __init__(self, n_observations, means: np.ndarray = None, covariance_matrix: np.ndarray = None):
        # Flag value that indicates function use, not method use
        # Therefore, we set a flag and immediately return
        if n_observations == -1:
            self.initialized = False
            return
        self.initialized = True
        self.covariance_matrix = covariance_matrix
        self.means = means

        self.n = n_observations
        n_features = self.n
        if means is not None:
            n_features = len(means)
        if covariance_matrix is not None:
            n_features = len(covariance_matrix)

        if covariance_matrix is None:
            self.covariance_matrix = np.eye(n_features)
        if means is None:
            self.means = np.zeros(n_features)

        self.df = self._generate_data()

    def _generate_data(self):
        result = np.random.multivariate_normal(self.means,
                                               self.covariance_matrix,
                                               size=self.n)
        return pd.DataFrame(result)

    def _check_initialized(self, df):
        """
        Checks to see if a Simulation object has been created. Since some methods are exposed as functions,
        this is meant to detect if a user calls a function without passing a df argument.
        The intended use of this module is to either (1) create a Simulation object if simulating data from scratch
        or (2) to use the exposed methods as functions and pass existing data as the df paramter
        :return:
        """
        if df is None:
            if not self.initialized:
                raise Exception(
                    'Please either (1) supply the `df` parameter if you have existing data or (2) initialize a '
                    'Simulation object.')
            return self.df  # User did not pass df and Simulation obj is initialized, this is the correct usage

        elif self.initialized:
            # df NOT None and Sim obj is initialized
            warnings.warn('Attempted to pass a `df` argument to a method of an initialized Simulation object. Will '
                          'default to the object\'s stored `df` attribute. If you wish to use a different `df` '
                          'please use the function instead of the method.')
            return self.df  # If user passes df even though Simulation was initialized
        return df

    def add_mean_drift(self, shifts_df, df=None):
        """
        *THIS METHOD IS DEPRECATED*
        Gradually shifts mean over time.
        This method uses the shifts_df parameter to give more control ovAer the rate at
        which the drift takes place. For example, a "rough" drift might include many
        small shifts in mean, along with a few larger shifts. Alternatively, to
        simulate a "smooth" drift, one can use the NumPy method
        `linspace(start, stop, num)` to build shift_df to create equal mean shifts
        over a specified interval starting at `start` ending at `end` with `num` steps.

        Args:
        df: DataFrame
            Data with features as columns and rows as observations.

        shifts_df: DataFrame
            Column names (as ints) correspond to feature indexes in df.
            Rows contain series of individual mean targets that are multiples of a
            column's standard deviation.
            i.e.
            --------------
            |   1   |   3  |
            |--------------|
            | 3.11  | 1.25 |
            | 3.21  | 1.30 |
            | 5.29  | 1.35 |
            --------------
            This data frame results in feature 1's mean being shifted to 3.11 sd's
            then 3.21 sd's then 5.29 sd's each by adding 1 observation to feature 1.
            Feature 3 is shifted to 1.25 sd's then 1.30 sd's, then 1.35 sd's each by
            adding 1 observation to feature 3.
        """
        df = self._check_initialized(df)

        local_df = df.copy(deep=True)
        feature_indexes = shifts_df.columns.values

        n_observations = local_df.shape[0]

        def val_needed_for_shift(data_list, target_mean, sd, mean):
            # print(f'Calculated mean target: mean+target*sd = {np.mean(data_list)}\
            # +{target_mean}*{sd}={np.mean(data_list)+sd*target_mean}\n')
            target_mean = mean + sd * target_mean
            needed_val = (len(data_list) + 1) * target_mean - sum(data_list)
            return needed_val

        sd_df = local_df[feature_indexes].std(
        )  # Save original std dev of columns
        means_df = local_df[feature_indexes].mean(
        )  # Save original means of columns
        for index, row in shifts_df.iterrows(
        ):  # iterate over rows of shift_df
            column_data = local_df[
                feature_indexes]  # extract feature columns to be shifted
            # Calculate size of point to add to induce desired mean shift
            # feature_means = column_data.mean()
            new_obs_list = []
            i = 0
            for col in column_data.columns:
                target_mean = row[row.keys()[i]]
                orig_sd = sd_df[row.keys()
                                [i]]  # original std dev for current column
                orig_mean = means_df[row.keys()
                                     [i]]  # original mean for current column
                v = val_needed_for_shift(column_data[col], target_mean,
                                         orig_sd, orig_mean)
                new_obs_list.append(v)
                i += 1
            # Add new points to df
            new_row = local_df.mean(
            )  # unaffected features will gain a mean point
            new_row[feature_indexes] = new_obs_list
            local_df = local_df.append(new_row, ignore_index=True)
        if self.initialized:
            self.df = local_df
        return local_df

    def add_anomalies(self, anomalies_df, df=None):
        """
        Adds anomalous points to dataframe

        Args:
        df: DataFrame
            Data with features as columns and rows as observations.

        anomalies_df: DataFrame
            Column names (as ints) correspond to feature indexes in df.
            Rows contain series of magnitudes as multiples of standard deviation.

            The standard deviation is calculated on the original df before the points
            are added (not recalculated after each point). This is done so that the
            size of anomalous points can easily be compared at function call. The
            standard deviation will only be recalculated on function call.

            Unaffected features gain 1 observation equal to the mean.
            i.e.
            ---------
            |  1  | 3 |
            |---------|
            | -4  | 3 |
            | -6  | 4 |
            |  6  | 5 |
            ---------
            This data frame results in feature 1 gaining 3 points: one -4*sd away from
            the mean, one -6*sd away from the mean, and one 6*sd away from the mean.
            Feature 3 gains 3 points: one 3*sd away from the mean, one 4*sd away from
            the mean, and one 5*sd away from the mean.
            All other features gain 1 observation equal to the mean.
        """
        df = self._check_initialized(df)

        local_df = df.copy(deep=True)
        feature_indexes = anomalies_df.columns.values

        n_observations = local_df.shape[0]

        # iterate over rows of shift_df
        for index, row in anomalies_df.iterrows():
            # Add new points to df
            new_row = local_df.mean(
            )  # unaffected features will gain a mean point
            new_row[feature_indexes] = local_df[feature_indexes].std() * row
            local_df = local_df.append(new_row, ignore_index=True)

        if self.initialized:
            self.df = local_df
        return local_df

    def add_gaussian_observations(self, summary_df, feature_index, df=None, visualize=False, append=False):
        """
        Args:
        summary_df:
            Contains mean and standard deviation of gaussian distribution being added to
            a feature.
            In practice, means should be calculated as a ratio of the standard deviation
            before being passed to this method.
            i.e.
            ----------------------
            |  mean |  sd  | n_obs |
            |--------------|-------|
            |  2.3  |  1.2 |   10  |
            |   0   |  1.3 |   20  |
            ----------------------
        """
        df = self._check_initialized(df)

        local_df = df.copy(deep=True)
        new_data = None
        for index, row in summary_df.iterrows(
        ):  # iterate over rows of summary_df
            mean = row['mean']
            sd = row['sd']
            n = int(row['n_obs'])
            d = np.random.normal(mean, sd, n)
            if new_data is None:
                new_data = d
            else:
                new_data = np.concatenate((new_data, d))

        total_n_obs = summary_df['n_obs'].sum()
        if append:  # Add new points by appending to df
            # Unaffected features gain gaussian obs w/ sd, mean estimated from sample
            temp_df = pd.DataFrame(columns=local_df.columns)
            mean_list = local_df.mean()
            sd_list = local_df.std()
            for mean, sd, col in zip(mean_list, sd_list, temp_df.columns):
                d = np.random.normal(mean, sd, total_n_obs)
                temp_df[col] = d
            # Append gaussian obs to local_df, affected feature will later be overwritten
            local_df = pd.concat([local_df, temp_df], ignore_index=True)

        # Add new points to df by overwriting
        local_df.loc[len(local_df) - total_n_obs:, feature_index] = new_data

        if visualize:
            import pyplot_themes as themes
            import seaborn as sns
            from matplotlib.pyplot import show, figure
            themes.theme_ggplot2()
            figure(figsize=(10, 7))
            normal_color = 'royalblue'
            anomalous_color = 'orangered'
            added_idx = len(local_df)-total_n_obs
            p = sns.lineplot(data=local_df[feature_index][:(added_idx+1)], color=normal_color)
            sns.lineplot(data=local_df[feature_index][added_idx :], ax=p, color=anomalous_color) # plot added obs
            sns.scatterplot(data=local_df[feature_index][added_idx : ], ax=p, color='black') # add dots to all new data
            p.set_xlabel('Observation')
            p.set_ylabel(f'Feature {feature_index}')
            p.set_title(f'Feature {feature_index} with Added Observations')
            # plt.savefig(data_dir+'step shift_plot1.png', dpi=300)
            show()

        if self.initialized:
            self.df = local_df
            return
        return local_df

    def get_data(self):
        return self.df


# Exposed methods that can be used as functions if the user does not use a Simulation object
# Choosing n=-1 is flag value (IGNORE: since only generate_data uses n and generate_data will not be exposed as a
# function, only as a method)
add_gaussian_observations = Simulation(-1).add_gaussian_observations




import matplotlib.pyplot as plt
import pyplot_themes as themes
import seaborn as sns
from mase.DataSimulation import *


def run_demo():
    # demo1()

    # Generate data with matching mean and correlation structure provided
    # means = summary_df[['mean']].to_numpy()
    # means = means.reshape(1, -1)[0]
    #
    # sim_data = generate_data(np.zeros(56), 200)
    # sim_df = pd.DataFrame(sim_data)

    # anomaly_demo()
    # new_df = demo_step_shift()
    # new_df = demo_smooth_shift(new_df)
    # new_df.to_csv(data_dir+'sim_data_1smooth_1step.csv', index=False)
    cov = np.eye(5)  # 5 independent features all with 0 mean
    sim = Simulation(100, covariance_matrix=cov)  # 100 observations
    summary_df = pd.DataFrame()
    summary_df['mean'] = [3, 0]
    summary_df['sd'] = [1, 1]
    summary_df['n_obs'] = [20, 10]
    feature_index = 0
    d = sim.get_data()
    print(d)
    sim.add_gaussian_observations(summary_df, feature_index, visualize=True)

    # Add obs to existing df
    # add_gaussian_observations(summary_df, feature_index, df=d, visualize=True, append=True)


def demo1():
    # Generate data with matching mean and correlation structure provided
    means = summary_df[['mean']].to_numpy()
    means = means.reshape(1, -1)[0]

    sim_data = generate_data(np.zeros(56), 100)
    sim_df = pd.DataFrame(sim_data)

    # Demonstrate mean drift
    # Feature 2 will undergo a smooth mean drift from 0.1sd to 1sd in 10 observations
    feature2_drift = list(np.linspace(0,1,num=20)) # 10 obs evenly spaced on []
    # Feature 6 will undergo a custom mean drift from 0.1 to 1 in 10 observations
    # feature6_drift = [0.1, 0.15, 0.15, 0.21, 0.3, 0.45, 0.65, 0.7, 0.5, 1]
    feature6_drift =feature2_drift

    shifts_df = pd.DataFrame(columns=[2,6])
    shifts_df[2] = feature2_drift
    shifts_df[6] = feature6_drift
    drifted_df = add_mean_drift(sim_df, shifts_df)
    print(f'Means before drift:\n{sim_df.mean()[[2,6]]}\n')
    print(f'Means after drift:\n{drifted_df.mean()[[2,6]]}\n')

    # Generate Plots to Demonstrate Effects
    sns.set()
    plt.rcParams['lines.linewidth'] = 3
    title_fontsize = 15
    tick_size = 12
    ax_font = {'weight':'semibold', 'size': 14}
    fig = plt.figure(figsize=(16,16))
    # Check feature 2
    ax = sns.lineplot(data=sim_df[2], ax=fig.add_subplot(221))
    sns.lineplot(data=drifted_df[2][99:], ax=ax)
    sns.scatterplot(data=drifted_df[2][100:], ax=ax, color='red')
    # ax.set_ylim(-1,23)
    ax.set_xlabel('Observation')
    ax.set_ylabel('Feature 2')
    ax.set_title('Feature 2 Observations After Adding Mean Drift')
    # Check feature 6
    ax = sns.lineplot(data=sim_df[6], ax=fig.add_subplot(222))
    sns.lineplot(data=drifted_df[6][99:], ax=ax)
    sns.scatterplot(data=drifted_df[6][100:], ax=ax, color='red')
    # ax.set_ylim(-1,23)
    ax.set_xlabel('Observation')
    ax.set_ylabel('Feature 6')
    ax.set_title('Feature 6 Observations After Adding Mean Drift')

    # Check feature 2 rolling mean
    f2_rolling_mean = sim_df[2].expanding(min_periods=5).mean()
    f2_drifted_rolling_mean = drifted_df[2].expanding(min_periods=5).mean()

    ax = sns.lineplot(data=f2_rolling_mean, ax=fig.add_subplot(223))
    sns.lineplot(data=f2_drifted_rolling_mean[99:], ax=ax)
    sns.scatterplot(data=f2_drifted_rolling_mean[99:], ax=ax, color='red')
    # ax.set_ylim(-1.5,1.5)
    ax.set_xlabel('Observation')
    ax.set_ylabel('Feature 2')
    ax.set_title('Feature 2 Rolling Mean')
    # Check feature 6 rolling mean
    f6_rolling_mean = sim_df[6].expanding(min_periods=5).mean()
    f6_drifted_rolling_mean = drifted_df[6].expanding(min_periods=5).mean()

    ax = sns.lineplot(data=f6_rolling_mean, ax=fig.add_subplot(224))
    sns.lineplot(data=f6_drifted_rolling_mean[99:], ax=ax)
    sns.scatterplot(data=f6_drifted_rolling_mean[100:], ax=ax, color='red')
    # ax.set_ylim(-1.5,1.5)

    ax.set_xlabel('Observation')
    ax.set_ylabel('Feature 6')
    ax.set_title('Feature 6 Rolling Mean')
    plt.show()


def demo_step_shift():
    # Demonstrate adding gaussian observations
    sd = sim_df[3].std()
    mean = sim_df[3].mean()

    step_shift_mean = mean+2.5*sd
    step_shift_sd = sd

    feature3_means = [step_shift_mean, # mean for step shift data
                    mean] # return to "normal" mean
    feature3_sd = [step_shift_sd, # sd for step shift data
                sd] # return to "normal" sd
    # Overwrite last 50 observations with 30 anomalous and 20 "normal"
    n_anomalous = 30 # Generate 30 step shift observations
    n_normal = 20 # End with 20 "normal" observations
    n_total = n_anomalous+n_normal # total number of observations we are simulating/overwriting
    feature3_n_obs = [n_anomalous, n_normal]

    f3summary_df = pd.DataFrame()
    f3summary_df['mean'] = feature3_means
    f3summary_df['sd'] = feature3_sd
    f3summary_df['n_obs'] = feature3_n_obs
    new_df = add_gaussian_observations(sim_df, f3summary_df, 3)

    # Generate plots to demonstrate effects

    fig = plt.figure(figsize=(15,5))
    # Check feature 3
    normal_color = 'royalblue'
    anomalous_color = 'orangered'
    anomalous_idx = len(sim_df)-n_total
    normal_idx = len(sim_df)-n_normal
    p = sns.lineplot(data=new_df[3][:(anomalous_idx+1)], color=normal_color)
    sns.lineplot(data=new_df[3][anomalous_idx : normal_idx], ax=p, color=anomalous_color) # plot anomalies
    sns.lineplot(data=new_df[3][normal_idx-1: ], ax=p, color=normal_color) # plot "normal"
    sns.scatterplot(data=new_df[3][anomalous_idx : ], ax=p, color='black') # add dots to all new data
    p.set_xlabel('Observation')
    p.set_ylabel('Feature 3')
    p.set_title('Feature 3 with $\mu+2.5\sigma$ Step Shift')
    # plt.savefig(data_dir+'step shift_plot1.png', dpi=300)
    plt.show()

    # Check feature 3 rolling mean
    f3_rolling_mean = sim_df[3].expanding(min_periods=5).mean()
    f3_new_rolling_mean = new_df[3].expanding(min_periods=5).mean()

    p2 = sns.lineplot(data=f3_rolling_mean[:anomalous_idx])
    sns.lineplot(data=f3_new_rolling_mean[anomalous_idx:], ax=p2)
    # ax.set_ylim(-1.5,1.5)
    p2.set_xlabel('Observation')
    p2.set_ylabel('Feature 3')
    p2.set_title('Feature 3 Rolling Mean')
    plt.show()
    return new_df


def demo_smooth_shift(df):
    # Demonstrate adding gaussian observations
    sd = sim_df[5].std()
    mean = sim_df[5].mean()

    n_means = 10 # will use 10 subtle mean changes
    feature5_means = mean+sd*np.linspace(0, 2, n_means)
    feature5_sd = list(np.repeat(sd, n_means))
    # Overwrite last 50 observations with 30 anomalous and 20 "normal"
    obs_per_shift = 5 # 5 observations per intermediate means
    n_total = n_means*obs_per_shift # total number of observations we are simulating/overwriting
    feature3_n_obs = np.repeat(obs_per_shift, n_means)

    f5summary_df = pd.DataFrame()
    f5summary_df['mean'] = feature5_means
    f5summary_df['sd'] = feature5_sd
    f5summary_df['n_obs'] = obs_per_shift
    new_df = add_gaussian_observations(df, f5summary_df, 5)

    # Generate plots to demonstrate effects
    themes.theme_ggplot2()
    fig = plt.figure(figsize=(15,5))
    # Check feature 3
    normal_color = 'royalblue'
    anomalous_color = 'orangered'
    anomalous_idx = len(sim_df)-n_total
    p = sns.lineplot(data=new_df[5][:(anomalous_idx+1)], color=normal_color)
    sns.lineplot(data=new_df[5][anomalous_idx : ], ax=p, color=anomalous_color) # plot anomalies
    sns.scatterplot(data=new_df[5][anomalous_idx : ], ax=p, color='black') # add dots to all new data
    p.set_xlabel('Observation')
    p.set_ylabel('Feature 5')
    p.set_title('Feature 5 with $\mu+2\sigma$ Smooth Shift')
    # plt.savefig(data_dir+'step shift_plot1.png', dpi=300)
    plt.show()

    # Check feature 5 rolling mean
    f5_rolling_mean = sim_df[5].expanding(min_periods=5).mean()
    f5_new_rolling_mean = new_df[5].expanding(min_periods=5).mean()

    p2 = sns.lineplot(data=f5_rolling_mean[:anomalous_idx])
    sns.lineplot(data=f5_new_rolling_mean[anomalous_idx:], ax=p2)
    # ax.set_ylim(-1.5,1.5)
    p2.set_xlabel('Observation')
    p2.set_ylabel('Feature 5')
    p2.set_title('Feature 5 Rolling Mean')
    plt.show()
    return new_df


def anomaly_demo():
    # Demonstrate adding anomalies
    anomalies_df = pd.DataFrame([[-4,3],
                                 [-6,4],
                                 [6,5]], columns=[1,3] )
    anomalous_df = add_anomalies(sim_df, anomalies_df)

    # Generate plots to demonstrate effects
    fig = plt.figure(figsize=(16,8))
    # Check feature 1
    ax = sns.lineplot(data=sim_df[1], ax=fig.add_subplot(121))
    sns.lineplot(data=anomalous_df[1][99:], ax=ax)
    sns.scatterplot(data=anomalous_df[1][100:], ax=ax, color='red')
    ax.set_xlabel('Observation')
    ax.set_ylabel('Feature 1')
    ax.set_title('Feature 1 with Anomalies Added')
    # Check feature 3
    ax = sns.lineplot(data=sim_df[3], ax=fig.add_subplot(122))
    sns.lineplot(data=anomalous_df[3][99:], ax=ax)
    sns.scatterplot(data=anomalous_df[3][100:], ax=ax, color='red')
    ax.set_xlabel('Observation')
    ax.set_ylabel('Feature 3')
    ax.set_title('Feature 3 with Anomalies Added')
    plt.show()


if __name__ == '__main__':
    run_demo()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="mase",
    version="0.0.5",
    author="Marco Christiani",
    author_email="mchristiani2017@gmail.com",
    description="A data simulation engine for multivariate time series data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Marco-Christiani/MASE",
    packages=setuptools.find_packages(exclude=["tests"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'numpy',
        'pandas',
        'matplotlib',
        'pyplot-themes',
        'seaborn',
    ]
)